% Soe San Win & David Truong
% MVDR Beamformer
% 1st-September-2009

% Systolic Implementation for MVDR Beamformer
% Comparison between One-step and Pre/Post Matrices

% Status: 2nd-September-2009 
% Work as exactly as stated in the theory.
% DONE !!!!!

%% Clear
clear all; clc; close all;

%% Simulation Parameters
N = 5;                                                                     % Number of Antennas
lambda = sqrt(1);                                                          % Exponential Weight (Forgetful Vector) 
M = 200;                                                                   % Number of Snapshot Iterations
n = (0:N-1)';                                                              % Indices for Steering Vector Calculation 

%% Variables
% Plot
sineTheta = (-1:(1/100):1)';                                               % Range of Plot Angles
plot_steer = [ones(201,1) exp(-1i*sineTheta*pi) exp(-2*1i*sineTheta*pi) exp(-3*1i*sineTheta*pi) exp(-4*1i*sineTheta*pi)];  % Steering Vector for Plotting

% Results
weights = zeros(N,M);                                                      % Weight matrix: Rows are the individual antennas & Columns are the time-steps 
weight_gr = zeros(N,M);
weight_lin = zeros(M,N);
post = zeros(N+2,N+1);
aux_gr = zeros(N,1);

% Inputs
z_input = 0;

%% Signal Parameters (Assume that the antennas are spaced wavelength/2 apart) 
SNR = 10;                                                                  % Signal-2-Noise Ratio (desired signal)
INR = 40;                                                                  % Interference-2-Noise Ratio (interference)
A_desired = sqrt(10^(SNR/10));                                             % Reverse Engineer --> SNR = 10log10(Desired_Amplitude/Noise_Amplitude)^2
A_interf = sqrt(10^(INR/10));                                              % Reverse Engineer --> INR = 10log10(Interference/Noise_Amplitude)^2
de_angle = -0.50*pi;                                                       % Angle of Desired Incidence == theta0 --> Desired Electric Angle: de_angle == theta0*pi
ine_angle = 0.10*pi;                                                       % Angle of Interfering Incidence == theta1 --> % Interfering Electrc Angle: ine_angle = theta1*pi

%% Input Vector Synthesis
U = zeros(N,M);                                                            % U is for calculating the Controlled Case                                                                               
buffer = zeros(M,N);                                                       % Buffer for stream of input samples 
steer = exp(-1i*n*de_angle);                                               % Steering Vector

for w = 1:M                                                                % Generate Signal Response
    noise = rand(N,1) + 1i* randn(N,1);
    psi = 2*pi*rand(N,1);
    desired = A_desired*exp(1i*(n-w)*de_angle);                            % Desired Sample
    interf = A_interf*exp(1i*(n-w)*ine_angle);                             % Interfering Sample
    buffer(w,:) = desired + interf + noise + psi;
end

for p=1:N
U(p,:) = buffer(:,p);                                                      % Initialize U 
end

%% Controlled Case Calculation
Phi_ctrl = U*U';
Phihalf_ctrl = chol(Phi_ctrl, 'lower');
Phihalf_inv_ctrl = inv(Phihalf_ctrl);
PhihalfH_inv_ctrl = Phihalf_inv_ctrl'; 
aux_ctrl = Phihalf_inv_ctrl*steer; 
up_ctrl = PhihalfH_inv_ctrl*aux_ctrl;  
bottom_ctrl = aux_ctrl'*aux_ctrl;
weight_ctrl = up_ctrl / bottom_ctrl;
% Calculate for the Plot Value
C_Y = weight_ctrl.'*plot_steer.';

%% Calculate for the Givens Rotation Simulation
Phihalf = eye(N,N);
aux_her = steer';
ident = eye(N+1,N+1);

for o=1:M
    input = U(:,o);
    pre = [lambda.*Phihalf input; lambda.*aux_her 0; zeros(1,N) 1];
    
    b_pre = pre;
    for l=1:N
        pre_temp = pre;
        f = pre_temp(l,l);
        g = pre_temp(l,(N+1));
        r = sqrt(abs(f)^2+abs(g)^2);
        
        if ((f ~= 0) && (g == 0))
            c = 1;
            s = 0;
        else
            c = abs(f)/r;
            s = g'/r;
        end
      
        gr = ident;
        gr(l,l) = c;
        gr(l,(N+1)) = -s';
        gr((N+1),l) = s;
        gr((N+1),(N+1)) = c;
        pre = pre_temp*gr;
    end
    post = pre;
    
    % Updating Values
    Phihalf = post(1:N,1:N);
    aux_her = post((N+1),(1:N));

    PhihalfH = Phihalf';
    aux_sqr = aux_her*aux_her';
    scaled_aux = (aux_her')./aux_sqr;
     
    % Calculate Weights using Linear Section
    for lindex=N:-1:1
        zbuff1 = 0;
        zbuff2 = 0;
        zbuff3 = 0;
        zbuff4 = 0;
        zbuff5 = 0;
        if (lindex == 5)
            [weight_lin(o,lindex)] =  linear_b(z_input,scaled_aux(lindex),PhihalfH(lindex,lindex));
        elseif (lindex == 4)
            [zbuff1] = linear_in(z_input,weight_lin(o,lindex+1),PhihalfH(lindex,lindex+1));
            [weight_lin(o,lindex)] = linear_b(zbuff1,scaled_aux(lindex),PhihalfH(lindex,lindex));
        elseif(lindex == 3)
            [zbuff1] = linear_in(z_input,weight_lin(o,lindex+2),PhihalfH(lindex,lindex+2));
            [zbuff2] = linear_in(zbuff1,weight_lin(o,lindex+1),PhihalfH(lindex,lindex+1));
            [weight_lin(o,lindex)] = linear_b(zbuff2,scaled_aux(lindex),PhihalfH(lindex,lindex));
        elseif(lindex == 2)
            [zbuff1] = linear_in(z_input,weight_lin(o,lindex+3),PhihalfH(lindex,lindex+3));
            [zbuff2] = linear_in(zbuff1,weight_lin(o,lindex+2),PhihalfH(lindex,lindex+2));
            [zbuff3] = linear_in(zbuff2,weight_lin(o,lindex+1),PhihalfH(lindex,lindex+1));
            [weight_lin(o,lindex)] = linear_b(zbuff3,scaled_aux(lindex),PhihalfH(lindex,lindex));
        elseif(lindex == 1)
            [zbuff1] = linear_in(z_input,weight_lin(o,lindex+4),PhihalfH(lindex,lindex+4));
            [zbuff2] = linear_in(zbuff1,weight_lin(o,lindex+3),PhihalfH(lindex,lindex+3));
            [zbuff3] = linear_in(zbuff2,weight_lin(o,lindex+2),PhihalfH(lindex,lindex+2));
            [zbuff4] = linear_in(zbuff3,weight_lin(o,lindex+1),PhihalfH(lindex,lindex+1));
            [weight_lin(o,lindex)] = linear_b(zbuff4,scaled_aux(lindex),PhihalfH(lindex,lindex));
        else
            zbuff1 = 0;
            zbuff2 = 0;
            zbuff3 = 0;
            zbuff4 = 0;
            zbuff5 = 0;
        end
    end
end
%Calculate for the Plot Value
Y = weight_lin(M,:)*plot_steer.';

%% Plotting
figure(1);
hold on;
plot(sineTheta,20*log10(abs(C_Y)),'b*');
plot(sineTheta,20*log10(abs(Y)),'r');
hold off;
grid;
title('Pre/Post Array Response');
xlabel('sine(angle)');
ylabel('Response (dB)');