% Soe San Win
% Internal Cell
% 25-June-2009

% This function will simulate the internal cell in the systolic array 
% structure. There are two types of internal cells as follows: 

% The first type will store the lower triangular matrix (Phi_half) element
% and perform the Given's Rotation to the incoming samples to cancel out
% the upper triangular array of the pre-array which is of the following 
% structure:
% 
% |  sqrt(lamda*Phi_half(previous-state))         u(current_state)   |
% |                                                                  |
% |                                                                  | 
% |  sqrt(lambda)*p(previous_state)'                      0          |
% |                                                                  | 
% |                                                                  |
% |        zero_vector'                                   1          |
% |                                                                  |
% Inputs: Samples (u) and Cosine (c) and Sine (s) of Rotation Angle and
% corresponding Lower Triangular Correlation Matrix (Phi_half) and 
% Exponential Weighting Factor (delta)
% Outputs: Rotated Input Samples (r_out)


% The second type will store the auxillary vector (aux) element and 
% perform the Given's Rotation done above.
% Inputs: Desired Response Vector (desired) (zero vector for MVDR problem)
% Outputs: 
 
function[u_out Phi c_out s_out] = internal(u,c,s,Phi,lambda)
    u_out = (c*u) - (s'*lambda*Phi);
    Phi = (s*u) + (c*lambda*Phi);
    c_out = c;
    s_out = s;
end 