% Soe San Win
% MVDR Beamformer
% 1-September-2009

% Systolic Implementation for MVDR Beamformer 
% Comparison between One-Step Solution and Systolic Implementation 

% Status: 3rd-September-2009
% Figure out the timing of calculating the Auxiliary Matrix, Weights and
% Phi. Plot it and verify it. 

%% Clear
clear all; clc; close all;

%% Simulation Parameters
N = 3;                                                                     % Number of Antennas
lambda = sqrt(1);                                                          % Exponential Weight (Forgetful Vector) 
M = 100;                                                                   % Number of Snapshot Iterations
n = (0:N-1)';                                                              % Indices for Steering Vector Calculation 
Scaler = 2^(-5);
%% Signal Parameters (Assume that the antennas are spaced wavelength/2 apart) 
SNR = 10;                                                                  % Signal-2-Noise Ratio (desired signal)
INR = 40;                                                                  % Interference-2-Noise Ratio (interference)
A_desired = sqrt(10^(SNR/10));                                             % Reverse Engineer --> SNR = 10log10(Desired_Amplitude/Noise_Amplitude)^2
A_interf = sqrt(10^(INR/10));                                              % Reverse Engineer --> INR = 10log10(Interference/Noise_Amplitude)^2
de_angle = -0.00*pi;                                                       % Angle of Desired Incidence == theta0 --> Desired Electric Angle: de_angle == theta0*pi
ine_angle = 0.30*pi;                                                       % Angle of Interfering Incidence == theta1 --> % Interfering Electrc Angle: ine_angle = theta1*pi

U = zeros(N,M);                                                            % U is for calculating the Controlled Case                                                                               
A = zeros(N,M+(2*N));                                                      % A is Nx(M+(N-1)) matrix for Systolic Case
buffer = zeros(M,N);                                                       % Buffer for stream of input samples 
steer = exp(-1i*n*de_angle);                                               % Steering Vector
%% Variables
% Plot
sineTheta = (-1:(1/100):1)';                                               % Range of Plot Angles
plot_steer = [ones(201,1) exp(-1i*sineTheta*pi) exp(-2*1i*sineTheta*pi)];  % Steering Vector for Plotting

% Results
weights = zeros(M+(2*N),N);                                                       % Weight matrix: Rows are the individual antennas & Columns are the time-steps 
weights_final = zeros(M+(2*N)-10,N);

% Intermediate Values
Phi_11 = ones(1,M+(2*N));                                                  % Phi Values in the Systolic Cells for respective positions
Phi_22 = ones(1,M+(2*N));
Phi_33 = ones(1,M+(2*N));
Phi_21 = zeros(1,M+(2*N));  
Phi_31 = zeros(1,M+(2*N));
Phi_32 = zeros(1,M+(2*N));
cs1 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];                                 % Rotation parameters 
cs2 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];
cs3 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];
cs4 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];
cs5 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];
cs6 = [ones(1,M+(2*N)); zeros(1,M+(2*N))];
inter1 = zeros(1,M+(2*N));                                                 % Intermediate rotated input samples before they are zeroed out.
inter2 = zeros(1,M+(2*N));
inter3 = zeros(1,M+(2*N));
inter4 = zeros(1,M+(2*N));
inter5 = zeros(1,M+(2*N));
inter6 = zeros(1,1);
auxsys_1 = zeros(1,M+(2*N));                                               % Calculated Auxiliary Vector     
auxsys_2 = zeros(1,M+(2*N));
auxsys_3 = zeros(1,M+(2*N));
aux1 = steer(1)'*ones(1,M+(2*N));                                           % Calculated Auxiliary Vector     
aux2 = steer(2)'*ones(1,M+(2*N));
aux3 = steer(3)'*ones(1,M+(2*N));
aux_scale = zeros(1,M+(2*N));
z1buff1 = zeros(1,M+(2*N));                                                % Buffers for Weight Vector Calculation
z2buff1 = zeros(1,M+(2*N));
z2buff2 = zeros(1,M+(2*N));

% Inputs
input = zeros(N,1);                                                        % Vector to the Systoloc Array
z_input = 0;                                                               % Zero Input     

%% Input Vector Synthesis
for w = 1:M                                                                % Generate Signal Response
    noise = rand(N,1) + 1i* randn(N,1);
    psi = 2*pi*rand(N,1);
    desired = A_desired*exp(1i*(n-w)*de_angle);                            % Desired Sample
    interf = A_interf*exp(1i*(n-w)*ine_angle);                             % Interfering Sample
    buffer(w,:) = desired + interf + noise;
end

for p=1:N
U(p,:) = buffer(:,p);                                                      % Initialize U 
end

A(1,:) = [U(1,:) zeros(1,2*N)];                                            % Initialize A   
A(2,:) = [zeros(1,N-2) U(2,:) zeros(1,2*N-1)];
A(3,:) = [zeros(1,N-1) U(3,:) zeros(1,2*N-2)];


%% Controlled Case Calculation
Phi_ctrl = U*U';
Phihalf_ctrl = chol(Phi_ctrl, 'lower');
Phihalf_inv_ctrl = inv(Phihalf_ctrl);
PhihalfH_inv_ctrl = Phihalf_inv_ctrl'; 
aux_ctrl = Phihalf_inv_ctrl*steer; 
up_ctrl = PhihalfH_inv_ctrl*aux_ctrl;  
bottom_ctrl = aux_ctrl'*aux_ctrl;
weight_ctrl = up_ctrl / bottom_ctrl;
% Calculate for the Plot Value
C_Y = weight_ctrl.'*plot_steer.';

%% Systolic Array
for time_step=1:(M+(2*N)) 
    input = Scaler.*A(:,time_step);                                       % Set the input vector coming into the systolic array for the first time-step      
    
    % Systolic Calculations
    [cs1(1,time_step+1) cs1(2,time_step+1) Phi_11(time_step+1)] = boundary(input(1),Phi_11(time_step));                                                                         % Boundary Cell (0,0)     
    [inter1(time_step+1) Phi_21(time_step+1) cs2(1,time_step+1) cs2(2,time_step+1)] = internal(input(2),cs1(1,time_step),cs1(2,time_step),Phi_21(time_step),lambda);            % Internal Cell (0,1)
    [cs3(1,time_step+1) cs3(2,time_step+1) Phi_22(time_step+1)] = boundary(inter1(time_step),Phi_22(time_step));                                                                % Boundary Cell (1,1) 
    [inter2(time_step+1) Phi_31(time_step+1) cs4(1,time_step+1) cs4(2,time_step+1)] = internal(input(3),cs2(1,time_step),cs2(2,time_step),Phi_31(time_step),lambda);            % Internal Cell (0,2)
    [inter3(time_step+1) Phi_32(time_step+1) cs5(1,time_step+1) cs5(2,time_step+1)] = internal(inter2(time_step),cs3(1,time_step),cs3(2,time_step),Phi_32(time_step),lambda);   % Internal Cell (1,2)
    [cs6(1,time_step+1) cs6(2,time_step+1) Phi_33(time_step+1)] = boundary(inter3(time_step),Phi_33(time_step));                                                                % Boundary Cell (2,2)
    [inter4(time_step+1) aux1(time_step+1)] = auxiliary(z_input,cs4(1,time_step),cs4(2,time_step),aux1(time_step),lambda);                                                      % Aux Cell (1)
    [inter5(time_step+1) aux2(time_step+1)] = auxiliary(inter4(time_step),cs5(1,time_step),cs5(2,time_step),aux2(time_step),lambda);                                            % Aux Cell (2)
    [inter6 aux3(time_step+1)] = auxiliary(inter5(time_step),cs6(1,time_step),cs6(2,time_step),aux3(time_step),lambda);                                                         % Aux Cell (3)
        
    % Auxiliary Matrix Scaling Factor Calculation 
    if (time_step > 6)
        aux_scale(time_step+1) = (aux1(time_step-2)'*aux1(time_step-2)) + (aux2(time_step-2)'*aux2(time_step-2)) + (aux3(time_step)'*aux3(time_step)); 
    end
    % Calculate Weights using Linear Section. No time_step+1 for buffers
    if (time_step > 7) % t=8
        [weights(time_step+1,3)] =  linear_b(z_input,(aux3(time_step-1)'/aux_scale(time_step)),Phi_33(time_step-2)');
    end
    if (time_step > 8) % t=9
        [z1buff1(time_step)] = linear_in(z_input,weights(time_step,3),Phi_32(time_step-4)');
        [weights(time_step+1,2)] = linear_b(z1buff1(time_step),(aux2(time_step-4)'/aux_scale(time_step-1)),Phi_22(time_step-5)');
    end
    if (time_step > 9) % t=10
        [z2buff1(time_step)] = linear_in(z_input,weights(time_step-1,3),Phi_31(time_step-6)');
        [z2buff2(time_step)] = linear_in(z2buff1(time_step),weights(time_step,2),Phi_21(time_step-7)');
        [weights(time_step+1,1)] = linear_b(z2buff2(time_step),(aux1(time_step-7)'/aux_scale(time_step-2)),Phi_11(time_step-8)');
    end
    if (time_step >10) % t=11
        weights_final(time_step-10,1) = weights(time_step,1);
        weights_final(time_step-10,2) = weights(time_step-1,2);
        weights_final(time_step-10,3) = weights(time_step-2,3);
    end
end
    Y = weights_final(M+(2*N)-10,:)*plot_steer.';

%% Plotting 
figure(1);
hold on;
plot(sineTheta,20*log10(abs(C_Y)),'b');
plot(sineTheta,20*log10(abs(Y)),'r*');
hold off;
grid;  
title('Systolic Response');
xlabel('sine(angle)');
ylabel('Response (dB)');